### EXESOFT PYIGNITION ###
# Copyright David Barker 2010
#
# Particle and ParticleSource objects


import keyframes, interpolate, random, math, pygame


DRAWTYPE_POINT = 0
DRAWTYPE_CIRCLE = 1
DRAWTYPE_LINE = 2
DRAWTYPE_SCALELINE = 3
DRAWTYPE_BUBBLE = 4
DRAWTYPE_IMAGE = 5


class Particle:
	def __init__(self, parent, initpos, velocity, life, drawtype = 0, colour = (0, 0, 0), radius = 0.0, length = 0.0, image = None, keyframes = []):
		self.parent = parent
		self.pos = initpos
		self.velocity = velocity
		self.life = life
		self.colour = colour
		self.radius = radius
		self.length = length
		self.image = image
		self.drawtype = drawtype
		
		self.keyframes = []
		self.keyframes.extend(keyframes[:])
		self.curframe = 0
		
		self.alive = True
	
	def Update(self):
		self.pos = [self.pos[0] + self.velocity[0], self.pos[1] + self.velocity[1]]
		
		if self.curframe > self.life:
			self.alive = False
		
		else:
			self.colour = (self.parent.particlecache[self.curframe]['colour_r'], self.parent.particlecache[self.curframe]['colour_g'], self.parent.particlecache[self.curframe]['colour_b'])
			self.radius = self.parent.particlecache[self.curframe]['radius']
			self.length = self.parent.particlecache[self.curframe]['length']
			
			self.curframe = self.curframe + 1
	
	def Draw(self, display):
		if (self.pos[0] > 10000) or (self.pos[1] > 10000) or (self.pos[0] < -10000) or (self.pos[1] < -10000):
			return
		
		if self.drawtype == DRAWTYPE_POINT:  # Point
			pygame.draw.circle(display, self.colour, self.pos, 0)
			
		elif self.drawtype == DRAWTYPE_CIRCLE:  # Circle
			pygame.draw.circle(display, self.colour, self.pos, self.radius)
		
		elif self.drawtype == DRAWTYPE_LINE:
			if self.length == 0.0:
				pygame.draw.circle(display, self.colour, self.pos, 0)
			
			else:
				velocitymagoverlength = math.sqrt(self.velocity[0]**2 + self.velocity[1]**2) / self.length
				linevec = [(self.velocity[0] / velocitymagoverlength), (self.velocity[1] / velocitymagoverlength)]
				endpoint = [self.pos[0] + linevec[0], self.pos[1] + linevec[1]]
				pygame.draw.aaline(display, self.colour, self.pos, endpoint)
			
		elif self.drawtype == DRAWTYPE_SCALELINE:  # Scaling line (scales with velocity)
			endpoint = [self.pos[0] + self.velocity[0], self.pos[1] + self.velocity[1]]
			pygame.draw.aaline(display, self.colour, self.pos, endpoint)
			
		elif self.drawtype == DRAWTYPE_BUBBLE:  # Bubble
			if self.radius >= 1.0:
				pygame.draw.circle(display, self.colour, self.pos, self.radius, 1)
			else:  # Pygame won't draw circles with thickness < radius, so if radius is smaller than one don't bother trying to set thickness
				pygame.draw.circle(display, self.colour, self.pos, self.radius)
		
		elif self.drawtype == DRAWTYPE_IMAGE:  # Image
			size = self.image.get_size()
			display.blit(self.image, (self.pos[0] - size[1], self.pos[1] - size[1]))

	def CreateKeyframe(self, frame, colour = (None, None, None), radius = None, length = None):
		keyframes.CreateKeyframe(self.keyframes, frame, {'colour_r':colour[0], 'colour_g':colour[1], 'colour_b':colour[2], 'radius':radius, 'length':length})


class ParticleSource:
	def __init__(self, parenteffect, pos, initspeed, initdirection, initspeedrandrange, initdirectionrandrange, particlesperframe, particlelife, genspacing, drawtype = 0, colour = (0, 0, 0), radius = 0.0, length = 0.0, image = None):
		self.parenteffect = parenteffect
		self.pos = pos
		self.initspeed = initspeed
		self.initdirection = initdirection
		self.initspeedrandrange = initspeedrandrange
		self.initdirectionrandrange = initdirectionrandrange
		self.particlesperframe = particlesperframe
		self.particlelife = particlelife
		self.genspacing = genspacing
		self.colour = colour
		self.drawtype = drawtype
		self.radius = radius
		self.length = length
		self.image = image

		self.keyframes = []
		self.CreateKeyframe(0, self.pos, self.initspeed, self.initdirection, self.initspeedrandrange, self.initdirectionrandrange, self.particlesperframe, self.genspacing)
		self.particlekeyframes = []
		self.particlecache = []
		self.CreateParticleKeyframe(0, colour = self.colour, radius = self.radius, length = self.length)
		self.curframe = 0
	
	def Update(self):		
		newvars = interpolate.InterpolateKeyframes(self.curframe, {'pos_x':self.pos[0], 'pos_y':self.pos[1], 'initspeed':self.initspeed, 'initdirection':self.initdirection, 'initspeedrandrange':self.initspeedrandrange, 'initdirectionrandrange':self.initdirectionrandrange, 'particlesperframe':self.particlesperframe, 'genspacing':self.genspacing}, self.keyframes)
		self.pos = (newvars['pos_x'], newvars['pos_y'])
		self.initspeed = newvars['initspeed']
		self.initdirection = newvars['initdirection']
		self.initspeedrandrange = newvars['initspeedrandrange']
		self.initdirectionrandrange = newvars['initdirectionrandrange']
		self.particlesperframe = newvars['particlesperframe']
		self.genspacing = newvars['genspacing']
		
		particlesperframe = self.particlesperframe
		
		if (self.genspacing == 0) or ((self.curframe % self.genspacing) == 0):
			for i in range(0, particlesperframe):
				self.CreateParticle()
		
		self.curframe = self.curframe + 1
	
	def CreateParticle(self):
		if self.initspeedrandrange != 0.0:
			speed = self.initspeed + (float(random.randrange(int(-self.initspeedrandrange * 100.0), int(self.initspeedrandrange * 100.0))) / 100.0)
		else:
			speed = self.initspeed
		if self.initdirectionrandrange != 0.0:
			direction = self.initdirection + (float(random.randrange(int(-self.initdirectionrandrange * 100.0), int(self.initdirectionrandrange * 100.0))) / 100.0)
		else:
			direction = self.initdirection
		velocity = [speed * math.sin(direction), -speed * math.cos(direction)]
		newparticle = Particle(self, initpos = self.pos, velocity = velocity, life = self.particlelife, drawtype = self.drawtype, colour = self.colour, radius = self.radius, length = self.length, image = self.image, keyframes = self.particlekeyframes)
		self.parenteffect.AddParticle(newparticle)

	def CreateKeyframe(self, frame, pos = (None, None), initspeed = None, initdirection = None, initspeedrandrange = None, initdirectionrandrange = None, particlesperframe = None, genspacing = None):
		keyframes.CreateKeyframe(self.keyframes, frame, {'pos_x':pos[0], 'pos_y':pos[1], 'initspeed':initspeed, 'initdirection':initdirection, 'initspeedrandrange':initspeedrandrange, 'initdirectionrandrange':initdirectionrandrange, 'particlesperframe':particlesperframe, 'genspacing':genspacing})
	
	def CreateParticleKeyframe(self, frame, colour = (None, None, None), radius = None, length = None):
		keyframes.CreateKeyframe(self.particlekeyframes, frame, {'colour_r':colour[0], 'colour_g':colour[1], 'colour_b':colour[2], 'radius':radius, 'length':length})
		self.PreCalculateParticles()
	
	def PreCalculateParticles(self):
		self.particlecache = []  # Clear the cache
		
		# Interpolate the particle variables for each frame of its life
		for i in xrange(0, self.particlelife + 1):
			vars = interpolate.InterpolateKeyframes(i, {'colour_r':0, 'colour_g':0, 'colour_b':0, 'radius':0, 'length':0}, self.particlekeyframes)
			self.particlecache.append(vars)
	
	def SetPos(self, newpos):
		self.CreateKeyframe(self.curframe, pos = newpos)
	
	def SetInitSpeed(self, newinitspeed):
		self.CreateKeyframe(self.curframe, initspeed = newinitspeed)
	
	def SetInitDirection(self, newinitdirection):
		self.CreateKeyframe(self.curframe, initdirection = newinitdirection)
	
	def SetInitSpeedRandRange(self, newinitspeedrandrange):
		self.CreateKeyframe(self.curframe, initspeedrangrange = newinitspeedrandrange)
	
	def SetInitDirectionRandRange(self, newinitdirectionrandrange):
		self.CreateKeyframe(self.curframe, initdirectionrandrange = newinitdirectionrandrange)
	
	def SetParticlesPerFrame(self, newparticlesperframe):
		self.CreateKeyframe(self.curframe, particlesperframe = newparticlesperframe)
	
	def SetGenSpacing(self, newgenspacing):
		self.CreateKeyframe(self.curframe, genspacing = newgenspacing)
